import typing as t
from pathlib import Path

from atproto_core.nsid import NSID
from atproto_lexicon.models import (
    LexObject,
    LexRef,
    LexXrpcParameters,
    LexXrpcProcedure,
    LexXrpcQuery,
)

from atproto_codegen.consts import (
    DISCLAIMER,
    INPUT_DICT,
    INPUT_MODEL,
    OUTPUT_MODEL,
    PARAMS_DICT,
    PARAMS_MODEL,
)
from atproto_codegen.namespaces.builder import MethodInfo, ProcedureInfo, QueryInfo, RecordInfo, build_namespaces
from atproto_codegen.record_templates import (
    RECORD_CREATE_METHOD_TEMPLATE,
    RECORD_DELETE_METHOD_TEMPLATE,
    RECORD_GET_METHOD_TEMPLATE,
    RECORD_LIST_METHOD_TEMPLATE,
)
from atproto_codegen.utils import (
    _resolve_nsid_ref,
    convert_camel_case_to_snake_case,
    format_code,
    gen_description_by_camel_case_name,
    get_import_path,
    get_sync_async_keywords,
    join_code,
    sort_dict_by_key,
    write_code,
)
from atproto_codegen.utils import get_code_intent as _

_NAMESPACES_OUTPUT_DIR = Path(__file__).parent.parent.parent.joinpath('atproto_client', 'namespaces')

_NAMESPACES_SYNC_FILENAME = 'sync_ns.py'
_NAMESPACES_ASYNC_FILENAME = 'async_ns.py'

_NAMESPACE_SUFFIX = 'Namespace'
_RECORD_SUFFIX = 'Record'


def get_namespace_name(path_parts: t.List[str]) -> str:
    return ''.join([p.capitalize() for p in path_parts]) + _NAMESPACE_SUFFIX


def get_record_name(path_parts: t.List[str]) -> str:
    return ''.join([p.capitalize() for p in path_parts]) + _RECORD_SUFFIX


def _get_namespace_imports() -> str:
    lines = [
        DISCLAIMER,
        'import typing as t',
        '',
        'from atproto_client import models',
        'from atproto_client.models.utils import get_or_create, get_response_model',
        'from atproto_client.namespaces.base import AsyncRecordBase, AsyncNamespaceBase, NamespaceBase, RecordBase',
        '',
        'if t.TYPE_CHECKING:',
        f'{_(1)}from atproto_client.client.async_raw import AsyncClientRaw',
        f'{_(1)}from atproto_client.client.raw import ClientRaw',
    ]

    return join_code(lines)


def _get_namespace_class_def(name_parts: t.List[str], *, sync: bool) -> str:
    base_class = 'NamespaceBase' if sync else 'AsyncNamespaceBase'
    lines = [f'class {get_namespace_name(name_parts)}({base_class}):']

    return join_code(lines)


def _get_record_class_def(name_parts: t.List[str], *, sync: bool) -> str:
    base_class = 'RecordBase' if sync else 'AsyncRecordBase'
    lines = [f'class {get_record_name(name_parts)}({base_class}):']

    return join_code(lines)


def _get_init_method(sub_items: t.List[str], *, sync: bool, parent_nodes: t.List[str], for_record: bool = False) -> str:
    if not sub_items:
        return ''

    client_typehint = "'ClientRaw'" if sync else "'AsyncClientRaw'"
    lines = [f'{_(1)}def __init__(self, client: {client_typehint}) -> None:', f'{_(2)}super().__init__(client)']

    get_name = get_namespace_name
    if for_record:
        get_name = get_record_name

    for sub_item in sub_items:
        lines.append(f'{_(2)}self.{sub_item} = {get_name([*parent_nodes, sub_item])}(self._client)')

    return join_code(lines)


def _get_method_docstring(method_info: MethodInfo) -> str:
    method_desc = f'{gen_description_by_camel_case_name(method_info.name)}.'
    if method_info.definition.description:
        method_desc = method_info.definition.description
    if method_desc[-1] not in {'.', '?', '!'}:
        method_desc += '.'

    doc_string = [f'{_(2)}"""{method_desc}', '', f'{_(2)}Args:']

    presented_args = _get_namespace_method_signature_args_names(method_info.definition)
    if 'params' in presented_args:
        doc_string.append(f'{_(3)}params: Parameters.')
    if 'data_schema' in presented_args:
        doc_string.append(f'{_(3)}data: Input data.')
    if 'data_schema_ref' in presented_args:
        doc_string.append(f'{_(3)}data: Input data (reference).')
    if 'data_alias' in presented_args:
        doc_string.append(f'{_(3)}data: Input data alias.')

    doc_string.append(f'{_(3)}**kwargs: Arbitrary arguments to HTTP request.')
    doc_string.append('')

    doc_string.append(f'{_(2)}Returns:')

    return_type, __ = _get_namespace_method_return_type(method_info)

    return_type_desc = 'Output model'
    if return_type == 'bool':
        return_type_desc = 'Success status'

    doc_string.append(f'{_(3)}:obj:`{return_type}`: {return_type_desc}.')
    doc_string.append('')

    doc_string.append(f'{_(2)}Raises:')
    doc_string.append(f'{_(3)}:class:`atproto.exceptions.AtProtocolError`: Base exception.')

    doc_string.append(f'{_(2)}"""')
    doc_string.append('')

    return join_code(doc_string)


@t.overload
def _get_namespace_method_body(method_info: ProcedureInfo, *, sync: bool) -> str: ...


@t.overload
def _get_namespace_method_body(method_info: QueryInfo, *, sync: bool) -> str: ...


def _get_namespace_method_body(method_info: MethodInfo, *, sync: bool) -> str:
    d, c = get_sync_async_keywords(sync=sync)

    lines = [_get_method_docstring(method_info)]

    presented_args = _get_namespace_method_signature_args_names(method_info.definition)
    presented_args.remove('self')

    def _override_arg_line(name: str, model_name: str) -> str:
        model_path = f'models.{get_import_path(method_info.nsid)}.{model_name}'
        # TODO(MarshalX): fix return type in get_or_create instead of casting
        return f"{_(2)}{name}_model = t.cast('{model_path}', get_or_create({name}, {model_path}))"

    invoke_args = [f"'{method_info.nsid}'"]

    if 'params' in presented_args:
        invoke_args.append('params=params_model')
        lines.append(_override_arg_line('params', PARAMS_MODEL))
    if 'data_schema' in presented_args:
        invoke_args.append('data=data_model')
        lines.append(_override_arg_line('data', INPUT_MODEL))
    if 'data_schema_ref' in presented_args:  # special case
        invoke_args.append('data=data_model')
        model_path, __ = _resolve_nsid_ref(method_info.nsid, method_info.definition.input.schema_.ref)
        # FIXME(MarshalX): typically this refs will lead to base.ModelBase, not models.DataModelBase
        #  which breaks type hinting in invoke methods (data argument)
        lines.append(f"{_(2)}data_model = t.cast('{model_path}', get_or_create(data, {model_path}))")
    if 'data_alias' in presented_args:
        invoke_args.append('data=data')
    if 'input_encoding' in presented_args:
        input_encoding = method_info.definition.input.encoding
        invoke_args.append(f"input_encoding='{input_encoding}'")
    if 'output_encoding' in presented_args:
        output_encoding = method_info.definition.output.encoding
        invoke_args.append(f"output_encoding='{output_encoding}'")

    invoke_args.append('**kwargs')
    invoke_args_str = ', '.join(invoke_args)

    method_name = 'invoke_query'
    if isinstance(method_info.definition, LexXrpcProcedure):
        method_name = 'invoke_procedure'

    lines.append(f'{_(2)}response = {c}self._client.{method_name}({invoke_args_str})')

    return_type, __ = _get_namespace_method_return_type(method_info)
    lines.append(f'{_(2)}return get_response_model(response, {return_type})')

    return join_code(lines)


def _get_namespace_method_signature_arg(
    name: str, nsid: NSID, model_name: t.Union[t.List[str], str], *, optional: bool, alias: bool = False
) -> str:
    if alias:
        return f"{name}: 'models.{get_import_path(nsid)}.{model_name}'"

    default_value = ''
    type_hint = (
        f"t.Union[dict, 'models.{get_import_path(nsid)}.{model_name}']"
        if isinstance(model_name, str)
        else 't.Union[' + ', '.join(f'models.{get_import_path(nsid)}.{i}' for i in model_name) + ']'
    )
    if optional:
        type_hint = f't.Optional[{type_hint}]'
        default_value = ' = None'

    return f'{name}: {type_hint}{default_value}'


def _get_namespace_method_signature_args_names(definition: t.Union[LexXrpcProcedure, LexXrpcQuery]) -> t.Set[str]:
    args = {'self'}
    if definition.parameters:
        args.add('params')

    if isinstance(definition, LexXrpcProcedure) and definition.input:
        if definition.input.schema_ and isinstance(definition.input.schema_, LexObject):
            args.add('data_schema')
        elif definition.input.schema_ and isinstance(definition.input.schema_, LexRef):
            args.add('data_schema_ref')
        else:
            args.add('data_alias')

        if definition.input.encoding:
            args.add('input_encoding')

    if definition.output and definition.output.encoding:
        args.add('output_encoding')

    return args


def _get_namespace_method_signature_args(method_info: MethodInfo) -> str:
    args = ['self']
    optional_args = []

    def _add_arg(arg_def: str, *, optional: bool) -> None:
        if optional:
            optional_args.append(arg_def)
        else:
            args.append(arg_def)

    def is_optional_arg(lex_obj: t.Union[LexObject, LexXrpcParameters]) -> bool:
        return lex_obj.required is None or len(lex_obj.required) == 0

    if method_info.definition.parameters:
        params = method_info.definition.parameters
        is_optional = is_optional_arg(params)

        arg = _get_namespace_method_signature_arg(
            'params', method_info.nsid, [PARAMS_MODEL, PARAMS_DICT], optional=is_optional
        )
        _add_arg(arg, optional=is_optional)

    if isinstance(method_info, ProcedureInfo) and method_info.definition.input:
        schema = method_info.definition.input.schema_
        if schema:
            if isinstance(schema, LexObject):
                is_optional = is_optional_arg(schema)
                arg = _get_namespace_method_signature_arg(
                    'data', method_info.nsid, [INPUT_MODEL, INPUT_DICT], optional=is_optional
                )
                _add_arg(arg, optional=is_optional)
            elif isinstance(schema, LexRef):
                model_path, _ = _resolve_nsid_ref(method_info.nsid, schema.ref)
                # there is no info about the requirement of LexRef ??? why
                _add_arg(f'data: {model_path}', optional=False)
            else:
                raise ValueError(f'Bad type {type(schema)}')  # probably LexRefUnion
        else:
            arg = _get_namespace_method_signature_arg('data', method_info.nsid, INPUT_MODEL, optional=False, alias=True)
            _add_arg(arg, optional=False)

    args.extend(optional_args)
    args.append('**kwargs: t.Any')
    return ', '.join(args)


def _get_namespace_method_return_type(method_info: MethodInfo) -> t.Tuple[str, bool]:
    if method_info.definition.output and isinstance(method_info.definition.output.schema_, LexRef):
        ref_class, _ = _resolve_nsid_ref(method_info.nsid, method_info.definition.output.schema_.ref)
        return ref_class, True

    is_model = False
    return_type = 'bool'  # return success of response
    if method_info.definition.output:
        # example of methods without response: app.bsky.graph.muteActor, app.bsky.graph.muteActor
        is_model = True
        return_type = f'models.{get_import_path(method_info.nsid)}.{OUTPUT_MODEL}'

    return return_type, is_model


def _get_namespace_method_signature(method_info: MethodInfo, *, sync: bool) -> str:
    d, c = get_sync_async_keywords(sync=sync)

    name = convert_camel_case_to_snake_case(method_info.name)
    args = _get_namespace_method_signature_args(method_info)
    return_type, is_model = _get_namespace_method_return_type(method_info)

    if is_model:
        return_type = f"'{return_type}'"

    return f'{_(1)}{d}def {name}({args}) -> {return_type}:'


def _get_namespace_methods_block(methods_info: t.List[MethodInfo], sync: bool) -> str:
    lines = []

    methods_info.sort(key=lambda e: e.name)
    for method_info in methods_info:
        lines.append(_get_namespace_method_signature(method_info, sync=sync))
        lines.append(_get_namespace_method_body(method_info, sync=sync))

    return join_code(lines)


def _get_record_methods_block(record_info: RecordInfo, sync: bool) -> str:
    d, c = get_sync_async_keywords(sync=sync)

    lines = []

    for method_template in [
        RECORD_GET_METHOD_TEMPLATE,
        RECORD_LIST_METHOD_TEMPLATE,
        RECORD_CREATE_METHOD_TEMPLATE,
        RECORD_DELETE_METHOD_TEMPLATE,
    ]:
        lines.append(
            method_template.format(
                record_import=get_import_path(record_info.nsid), collection=record_info.nsid, d=d, c=c
            )
        )

    return join_code(lines)


def _get_record_class(records_info: t.List[RecordInfo], sync: bool) -> str:
    lines = []

    for record_info in records_info:
        lines.append(_get_record_class_def(record_info.nsid.segments, sync=sync))
        lines.append(_get_record_methods_block(record_info, sync=sync))

    return join_code(lines)


def _generate_namespace_in_output(
    namespace_tree: dict, output: t.List[str], *, sync: bool, parent_nodes: t.List[str]
) -> None:
    for node_name, sub_node in namespace_tree.items():
        nodes_path = [*parent_nodes, node_name]

        if isinstance(sub_node, dict):
            output.append(_get_namespace_class_def(nodes_path, sync=sync))

            sub_nodes_names = list(sort_dict_by_key(sub_node).keys())
            output.append(_get_init_method(sub_nodes_names, sync=sync, parent_nodes=nodes_path))

            _generate_namespace_in_output(sub_node, output, sync=sync, parent_nodes=nodes_path)

        if isinstance(sub_node, list):
            records = [info for info in sub_node if isinstance(info, RecordInfo)]
            methods = [info for info in sub_node if isinstance(info, (ProcedureInfo, QueryInfo))]

            output.append(_get_record_class(records, sync=sync))

            output.append(_get_namespace_class_def(nodes_path, sync=sync))

            record_names = sorted([record.name for record in records])
            output.append(_get_init_method(record_names, sync=sync, parent_nodes=nodes_path, for_record=True))

            output.append(_get_namespace_methods_block(methods, sync=sync))


def generate_namespaces(
    lexicon_dir: t.Optional[Path] = None,
    output_dir: t.Optional[Path] = None,
    async_filename: t.Optional[str] = None,
    sync_filename: t.Optional[str] = None,
) -> None:
    if not output_dir:
        output_dir = _NAMESPACES_OUTPUT_DIR
    if not async_filename:
        async_filename = _NAMESPACES_ASYNC_FILENAME
    if not sync_filename:
        sync_filename = _NAMESPACES_SYNC_FILENAME

    namespace_tree = build_namespaces(lexicon_dir)

    for sync in (True, False):
        generated_code_lines_buffer: t.List[str] = []
        _generate_namespace_in_output(namespace_tree, generated_code_lines_buffer, sync=sync, parent_nodes=[])

        code = join_code([_get_namespace_imports(), *generated_code_lines_buffer])

        filename = sync_filename if sync else async_filename
        filepath = output_dir.joinpath(filename)

        write_code(filepath, code)

    # TODO(MarshalX): generate ClientRaw as root of namespaces

    format_code(output_dir)



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2024 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t

import typing_extensions as te

from atproto_client.models import base


class Data(base.DataModelBase):
    """Input data model for :obj:`chat.bsky.convo.updateAllRead`."""

    status: t.Optional[t.Union[t.Literal['request'], t.Literal['accepted'], str]] = None  #: Status.


class DataDict(t.TypedDict):
    status: te.NotRequired[t.Optional[t.Union[t.Literal['request'], t.Literal['accepted'], str]]]  #: Status.


class Response(base.ResponseModelBase):
    """Output data model for :obj:`chat.bsky.convo.updateAllRead`."""

    updated_count: int  #: The count of updated convos.



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2024 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t

from pydantic import Field

from atproto_client.models import string_formats

if t.TYPE_CHECKING:
    from atproto_client import models
from atproto_client.models import base


class Member(base.ModelBase):
    """Definition model for :obj:`tools.ozone.team.defs`."""

    did: string_formats.Did  #: Did.
    role: t.Union[
        'models.ToolsOzoneTeamDefs.RoleAdmin',
        'models.ToolsOzoneTeamDefs.RoleModerator',
        'models.ToolsOzoneTeamDefs.RoleTriage',
        str,
    ]  #: Role.
    created_at: t.Optional[string_formats.DateTime] = None  #: Created at.
    disabled: t.Optional[bool] = None  #: Disabled.
    last_updated_by: t.Optional[str] = None  #: Last updated by.
    profile: t.Optional['models.AppBskyActorDefs.ProfileViewDetailed'] = None  #: Profile.
    updated_at: t.Optional[string_formats.DateTime] = None  #: Updated at.

    py_type: t.Literal['tools.ozone.team.defs#member'] = Field(
        default='tools.ozone.team.defs#member', alias='$type', frozen=True
    )


RoleAdmin = t.Literal[
    'tools.ozone.team.defs#roleAdmin'
]  #: Admin role. Highest level of access, can perform all actions.

RoleModerator = t.Literal['tools.ozone.team.defs#roleModerator']  #: Moderator role. Can perform most actions.

RoleTriage = t.Literal[
    'tools.ozone.team.defs#roleTriage'
]  #: Triage role. Mostly intended for monitoring and escalating issues.



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2024 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t

from pydantic import Field

from atproto_client.models import base


class AspectRatio(base.ModelBase):
    """Definition model for :obj:`app.bsky.embed.defs`. width:height represents an aspect ratio. It may be approximate, and may not correspond to absolute dimensions in any given unit."""

    height: int = Field(ge=1)  #: Height.
    width: int = Field(ge=1)  #: Width.

    py_type: t.Literal['app.bsky.embed.defs#aspectRatio'] = Field(
        default='app.bsky.embed.defs#aspectRatio', alias='$type', frozen=True
    )



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2024 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t

from pydantic import Field

from atproto_client.models import base, string_formats


class TemplateView(base.ModelBase):
    """Definition model for :obj:`tools.ozone.communication.defs`."""

    content_markdown: str  #: Subject of the message, used in emails.
    created_at: string_formats.DateTime  #: Created at.
    disabled: bool  #: Disabled.
    id: str  #: Id.
    last_updated_by: string_formats.Did  #: DID of the user who last updated the template.
    name: str  #: Name of the template.
    updated_at: string_formats.DateTime  #: Updated at.
    lang: t.Optional[string_formats.Language] = None  #: Message language.
    subject: t.Optional[str] = None  #: Content of the template, can contain markdown and variable placeholders.

    py_type: t.Literal['tools.ozone.communication.defs#templateView'] = Field(
        default='tools.ozone.communication.defs#templateView', alias='$type', frozen=True
    )



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2024 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t

import typing_extensions as te
from pydantic import Field

from atproto_client.models import string_formats

if t.TYPE_CHECKING:
    from atproto_client import models
    from atproto_client.models.unknown_type import UnknownType
from atproto_client.models import base


class Data(base.DataModelBase):
    """Input data model for :obj:`com.atproto.repo.applyWrites`."""

    repo: string_formats.AtIdentifier  #: The handle or DID of the repo (aka, current account).
    writes: t.List[
        te.Annotated[
            t.Union[
                'models.ComAtprotoRepoApplyWrites.Create',
                'models.ComAtprotoRepoApplyWrites.Update',
                'models.ComAtprotoRepoApplyWrites.Delete',
            ],
            Field(discriminator='py_type'),
        ]
    ]  #: Writes.
    swap_commit: t.Optional[string_formats.Cid] = (
        None  #: If provided, the entire operation will fail if the current repo commit CID does not match this value. Used to prevent conflicting repo mutations.
    )
    validate_: t.Optional[bool] = (
        None  #: Can be set to 'false' to skip Lexicon schema validation of record data across all operations, 'true' to require it, or leave unset to validate only for known Lexicons.
    )


class DataDict(t.TypedDict):
    repo: string_formats.AtIdentifier  #: The handle or DID of the repo (aka, current account).
    writes: t.List[
        te.Annotated[
            t.Union[
                'models.ComAtprotoRepoApplyWrites.Create',
                'models.ComAtprotoRepoApplyWrites.Update',
                'models.ComAtprotoRepoApplyWrites.Delete',
            ],
            Field(discriminator='py_type'),
        ]
    ]  #: Writes.
    swap_commit: te.NotRequired[
        t.Optional[string_formats.Cid]
    ]  #: If provided, the entire operation will fail if the current repo commit CID does not match this value. Used to prevent conflicting repo mutations.
    validate: te.NotRequired[
        t.Optional[bool]
    ]  #: Can be set to 'false' to skip Lexicon schema validation of record data across all operations, 'true' to require it, or leave unset to validate only for known Lexicons.


class Response(base.ResponseModelBase):
    """Output data model for :obj:`com.atproto.repo.applyWrites`."""

    commit: t.Optional['models.ComAtprotoRepoDefs.CommitMeta'] = None  #: Commit.
    results: t.Optional[
        t.List[
            te.Annotated[
                t.Union[
                    'models.ComAtprotoRepoApplyWrites.CreateResult',
                    'models.ComAtprotoRepoApplyWrites.UpdateResult',
                    'models.ComAtprotoRepoApplyWrites.DeleteResult',
                ],
                Field(discriminator='py_type'),
            ]
        ]
    ] = None  #: Results.


class Create(base.ModelBase):
    """Definition model for :obj:`com.atproto.repo.applyWrites`. Operation which creates a new record."""

    collection: string_formats.Nsid  #: Collection.
    value: 'UnknownType'  #: Value.
    rkey: t.Optional[string_formats.RecordKey] = Field(
        default=None, max_length=512
    )  #: NOTE: maxLength is redundant with record-key format. Keeping it temporarily to ensure backwards compatibility.

    py_type: t.Literal['com.atproto.repo.applyWrites#create'] = Field(
        default='com.atproto.repo.applyWrites#create', alias='$type', frozen=True
    )


class Update(base.ModelBase):
    """Definition model for :obj:`com.atproto.repo.applyWrites`. Operation which updates an existing record."""

    collection: string_formats.Nsid  #: Collection.
    rkey: string_formats.RecordKey  #: Rkey.
    value: 'UnknownType'  #: Value.

    py_type: t.Literal['com.atproto.repo.applyWrites#update'] = Field(
        default='com.atproto.repo.applyWrites#update', alias='$type', frozen=True
    )


class Delete(base.ModelBase):
    """Definition model for :obj:`com.atproto.repo.applyWrites`. Operation which deletes an existing record."""

    collection: string_formats.Nsid  #: Collection.
    rkey: string_formats.RecordKey  #: Rkey.

    py_type: t.Literal['com.atproto.repo.applyWrites#delete'] = Field(
        default='com.atproto.repo.applyWrites#delete', alias='$type', frozen=True
    )


class CreateResult(base.ModelBase):
    """Definition model for :obj:`com.atproto.repo.applyWrites`."""

    cid: string_formats.Cid  #: Cid.
    uri: string_formats.AtUri  #: Uri.
    validation_status: t.Optional[t.Union[t.Literal['valid'], t.Literal['unknown'], str]] = None  #: Validation status.

    py_type: t.Literal['com.atproto.repo.applyWrites#createResult'] = Field(
        default='com.atproto.repo.applyWrites#createResult', alias='$type', frozen=True
    )


class UpdateResult(base.ModelBase):
    """Definition model for :obj:`com.atproto.repo.applyWrites`."""

    cid: string_formats.Cid  #: Cid.
    uri: string_formats.AtUri  #: Uri.
    validation_status: t.Optional[t.Union[t.Literal['valid'], t.Literal['unknown'], str]] = None  #: Validation status.

    py_type: t.Literal['com.atproto.repo.applyWrites#updateResult'] = Field(
        default='com.atproto.repo.applyWrites#updateResult', alias='$type', frozen=True
    )


class DeleteResult(base.ModelBase):
    """Definition model for :obj:`com.atproto.repo.applyWrites`."""

    py_type: t.Literal['com.atproto.repo.applyWrites#deleteResult'] = Field(
        default='com.atproto.repo.applyWrites#deleteResult', alias='$type', frozen=True
    )



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2024 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t

import typing_extensions as te
from pydantic import Field

from atproto_client.models import string_formats

if t.TYPE_CHECKING:
    from atproto_client import models
from atproto_client.models import base


class Params(base.ParamsModelBase):
    """Parameters model for :obj:`app.bsky.labeler.getServices`."""

    dids: t.List[string_formats.Did]  #: Dids.
    detailed: t.Optional[bool] = False  #: Detailed.


class ParamsDict(t.TypedDict):
    dids: t.List[string_formats.Did]  #: Dids.
    detailed: te.NotRequired[t.Optional[bool]]  #: Detailed.


class Response(base.ResponseModelBase):
    """Output data model for :obj:`app.bsky.labeler.getServices`."""

    views: t.List[
        te.Annotated[
            t.Union['models.AppBskyLabelerDefs.LabelerView', 'models.AppBskyLabelerDefs.LabelerViewDetailed'],
            Field(discriminator='py_type'),
        ]
    ]  #: Views.



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2024 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t

from atproto_client.models import base


class Response(base.ResponseModelBase):
    """Output data model for :obj:`app.bsky.video.getUploadLimits`."""

    can_upload: bool  #: Can upload.
    error: t.Optional[str] = None  #: Error.
    message: t.Optional[str] = None  #: Message.
    remaining_daily_bytes: t.Optional[int] = None  #: Remaining daily bytes.
    remaining_daily_videos: t.Optional[int] = None  #: Remaining daily videos.



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2024 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t

from atproto_client.models import base, string_formats


class Params(base.ParamsModelBase):
    """Parameters model for :obj:`com.atproto.sync.getRepoStatus`."""

    did: string_formats.Did  #: The DID of the repo.


class ParamsDict(t.TypedDict):
    did: string_formats.Did  #: The DID of the repo.


class Response(base.ResponseModelBase):
    """Output data model for :obj:`com.atproto.sync.getRepoStatus`."""

    active: bool  #: Active.
    did: string_formats.Did  #: Did.
    rev: t.Optional[string_formats.Tid] = None  #: Optional field, the current rev of the repo, if active=true.
    status: t.Optional[
        t.Union[
            t.Literal['takendown'],
            t.Literal['suspended'],
            t.Literal['deleted'],
            t.Literal['deactivated'],
            t.Literal['desynchronized'],
            t.Literal['throttled'],
            str,
        ]
    ] = None  #: If active=false, this optional field indicates a possible reason for why the account is not active. If active=false and no status is supplied, then the host makes no claim for why the repository is no longer being hosted.



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2024 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t

import typing_extensions as te
from pydantic import Field

from atproto_client.models import base, string_formats


class Data(base.DataModelBase):
    """Input data model for :obj:`com.atproto.server.createAppPassword`."""

    name: str  #: A short name for the App Password, to help distinguish them.
    privileged: t.Optional[bool] = (
        None  #: If an app password has 'privileged' access to possibly sensitive account state. Meant for use with trusted clients.
    )


class DataDict(t.TypedDict):
    name: str  #: A short name for the App Password, to help distinguish them.
    privileged: te.NotRequired[
        t.Optional[bool]
    ]  #: If an app password has 'privileged' access to possibly sensitive account state. Meant for use with trusted clients.


class AppPassword(base.ModelBase):
    """Definition model for :obj:`com.atproto.server.createAppPassword`."""

    created_at: string_formats.DateTime  #: Created at.
    name: str  #: Name.
    password: str  #: Password.
    privileged: t.Optional[bool] = None  #: Privileged.

    py_type: t.Literal['com.atproto.server.createAppPassword#appPassword'] = Field(
        default='com.atproto.server.createAppPassword#appPassword', alias='$type', frozen=True
    )



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2024 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t

from pydantic import Field

from atproto_client.models import base, string_formats


class Data(base.DataModelBase):
    """Input data model for :obj:`tools.ozone.setting.removeOptions`."""

    keys: t.List[string_formats.Nsid] = Field(min_length=1, max_length=200)  #: Keys.
    scope: t.Union[t.Literal['instance'], t.Literal['personal'], str]  #: Scope.


class DataDict(t.TypedDict):
    keys: t.List[string_formats.Nsid]  #: Keys.
    scope: t.Union[t.Literal['instance'], t.Literal['personal'], str]  #: Scope.


class Response(base.ResponseModelBase):
    """Output data model for :obj:`tools.ozone.setting.removeOptions`."""



import typing as t
from threading import Lock

from atproto_core.uri import AtUri

from atproto_client import models
from atproto_client.client.methods_mixin import SessionMethodsMixin, TimeMethodsMixin
from atproto_client.client.methods_mixin.headers import HeadersConfigurationMethodsMixin
from atproto_client.client.methods_mixin.session import SessionDispatchMixin
from atproto_client.client.raw import ClientRaw
from atproto_client.client.session import Session, SessionEvent, SessionResponse
from atproto_client.exceptions import LoginRequiredError
from atproto_client.models.languages import DEFAULT_LANGUAGE_CODE1
from atproto_client.utils import TextBuilder

if t.TYPE_CHECKING:
    from atproto_client.client.base import InvokeType
    from atproto_client.request import Response


class Client(SessionDispatchMixin, SessionMethodsMixin, TimeMethodsMixin, HeadersConfigurationMethodsMixin, ClientRaw):
    """High-level client for XRPC of ATProto."""

    def __init__(self, base_url: t.Optional[str] = None, *args: t.Any, **kwargs: t.Any) -> None:
        super().__init__(base_url, *args, **kwargs)

        self._refresh_lock = Lock()

        self.me: t.Optional['models.AppBskyActorDefs.ProfileViewDetailed'] = None

    def _invoke(self, invoke_type: 'InvokeType', **kwargs: t.Any) -> 'Response':
        ignore_session_check = kwargs.pop('ignore_session_check', False)
        if ignore_session_check:
            return super()._invoke(invoke_type, **kwargs)

        with self._refresh_lock:
            if self._session and self._session.access_jwt and self._should_refresh_session():
                self._refresh_and_set_session()

        return super()._invoke(invoke_type, **kwargs)

    def _set_session(self, event: SessionEvent, session: SessionResponse) -> None:
        self._set_session_common(session, self._base_url)
        self._call_on_session_change_callbacks(event)

    def _get_and_set_session(
        self, login: str, password: str, auth_factor_token: t.Optional[str] = None
    ) -> 'models.ComAtprotoServerCreateSession.Response':
        session = self.com.atproto.server.create_session(
            models.ComAtprotoServerCreateSession.Data(
                identifier=login, password=password, auth_factor_token=auth_factor_token
            ),
            ignore_session_check=True,
        )
        self._set_session(SessionEvent.CREATE, session)
        return session

    def _refresh_and_set_session(self) -> 'models.ComAtprotoServerRefreshSession.Response':
        if not self._session or not self._session.refresh_jwt:
            raise LoginRequiredError

        refresh_session = self.com.atproto.server.refresh_session(
            headers=self._get_refresh_auth_headers(), ignore_session_check=True
        )
        self._set_session(SessionEvent.REFRESH, refresh_session)

        return refresh_session

    def _import_session_string(self, session_string: str) -> Session:
        import_session = Session.decode(session_string)
        self._set_session(SessionEvent.IMPORT, import_session)

        return import_session

    def login(
        self,
        login: t.Optional[str] = None,
        password: t.Optional[str] = None,
        session_string: t.Optional[str] = None,
        auth_factor_token: t.Optional[str] = None,
    ) -> 'models.AppBskyActorDefs.ProfileViewDetailed':
        """Authorize a client and get profile info.

        Args:
            login: Handle/username of the account.
            password: Main or app-specific password of the account.
            session_string: Session string (use :py:attr:`~export_session_string` to get it).
            auth_factor_token: Auth factor token (for Email 2FA).

        Note:
            Either `session_string` or `login` and `password` should be provided.

        Returns:
            :obj:`models.AppBskyActorDefs.ProfileViewDetailed`: Profile information.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """
        if session_string:
            session = self._import_session_string(session_string)
        elif login and password:
            session = self._get_and_set_session(login, password, auth_factor_token)
        else:
            raise ValueError('Either session_string or login and password should be provided.')

        self.me = self.app.bsky.actor.get_profile(models.AppBskyActorGetProfile.Params(actor=session.handle))
        return self.me

    def send_post(
        self,
        text: t.Union[str, TextBuilder],
        profile_identify: t.Optional[str] = None,
        reply_to: t.Optional['models.AppBskyFeedPost.ReplyRef'] = None,
        embed: t.Optional[
            t.Union[
                'models.AppBskyEmbedImages.Main',
                'models.AppBskyEmbedExternal.Main',
                'models.AppBskyEmbedRecord.Main',
                'models.AppBskyEmbedRecordWithMedia.Main',
                'models.AppBskyEmbedVideo.Main',
            ]
        ] = None,
        langs: t.Optional[t.List[str]] = None,
        facets: t.Optional[t.List['models.AppBskyRichtextFacet.Main']] = None,
    ) -> 'models.AppBskyFeedPost.CreateRecordResponse':
        """Send post.

        Note:
            If `profile_identify` is not provided will be sent to the current profile.

            The default language is ``en``.
            Available languages are defined in :py:mod:`atproto.xrpc_client.models.languages`.

        Args:
            text: Text of the post.
            profile_identify: Handle or DID. Where to send post.
            reply_to: Root and parent of the post to reply to.
            embed: Embed models that should be attached to the post.
            langs: List of used languages in the post.
            facets: List of facets (rich text items).

        Returns:
            :obj:`models.AppBskyFeedPost.CreateRecordResponse`: Reference to the created record.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """
        if isinstance(text, TextBuilder):
            facets = text.build_facets()
            text = text.build_text()

        repo = self.me and self.me.did
        if profile_identify:
            repo = profile_identify

        if not repo:
            raise LoginRequiredError

        if not langs:
            langs = [DEFAULT_LANGUAGE_CODE1]

        record = models.AppBskyFeedPost.Record(
            created_at=self.get_current_time_iso(),
            text=text,
            reply=reply_to,
            embed=embed,
            langs=langs,
            facets=facets,
        )
        return self.app.bsky.feed.post.create(repo, record)

    def delete_post(self, post_uri: str) -> bool:
        """Delete post.

        Args:
            post_uri: AT URI of the post.

        Returns:
            :obj:`bool`: Success status.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """
        uri = AtUri.from_str(post_uri)
        return self.app.bsky.feed.post.delete(uri.hostname, uri.rkey)

    def send_images(
        self,
        text: t.Union[str, TextBuilder],
        images: t.List[bytes],
        image_alts: t.Optional[t.List[str]] = None,
        profile_identify: t.Optional[str] = None,
        reply_to: t.Optional['models.AppBskyFeedPost.ReplyRef'] = None,
        langs: t.Optional[t.List[str]] = None,
        facets: t.Optional[t.List['models.AppBskyRichtextFacet.Main']] = None,
        image_aspect_ratios: t.Optional[t.List['models.AppBskyEmbedDefs.AspectRatio']] = None,
    ) -> 'models.AppBskyFeedPost.CreateRecordResponse':
        """Send post with multiple attached images (up to 4 images).

        Note:
            If `profile_identify` is not provided will be sent to the current profile.

        Args:
            text: Text of the post.
            images: List of binary images to attach. The length must be less than or equal to 4.
            image_alts: List of text version of the images.
                        The length should be shorter than or equal to the length of `images`.
            profile_identify: Handle or DID. Where to send post.
            reply_to: Root and parent of the post to reply to.
            langs: List of used languages in the post.
            facets: List of facets (rich text items).
            image_aspect_ratios: List of aspect ratios of the images.
                        The length should be shorter than or equal to the length of `images`.

        Returns:
            :obj:`models.AppBskyFeedPost.CreateRecordResponse`: Reference to the created record.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """
        if image_alts is None:
            image_alts = [''] * len(images)
        else:
            # padding with empty string if len is insufficient
            diff = len(images) - len(image_alts)
            image_alts = image_alts + [''] * diff  # [''] * (minus) => []

        if image_aspect_ratios is None:
            aligned_image_aspect_ratios = [None] * len(images)
        else:
            # padding with None if len is insufficient
            diff = len(images) - len(image_aspect_ratios)
            aligned_image_aspect_ratios = image_aspect_ratios + [None] * diff

        uploads = [self.upload_blob(image) for image in images]
        embed_images = [
            models.AppBskyEmbedImages.Image(alt=alt, image=upload.blob, aspect_ratio=aspect_ratio)
            for alt, upload, aspect_ratio in zip(image_alts, uploads, aligned_image_aspect_ratios)
        ]

        return self.send_post(
            text,
            profile_identify=profile_identify,
            reply_to=reply_to,
            embed=models.AppBskyEmbedImages.Main(images=embed_images),
            langs=langs,
            facets=facets,
        )

    def send_image(
        self,
        text: t.Union[str, TextBuilder],
        image: bytes,
        image_alt: str,
        profile_identify: t.Optional[str] = None,
        reply_to: t.Optional['models.AppBskyFeedPost.ReplyRef'] = None,
        langs: t.Optional[t.List[str]] = None,
        facets: t.Optional[t.List['models.AppBskyRichtextFacet.Main']] = None,
        image_aspect_ratio: t.Optional['models.AppBskyEmbedDefs.AspectRatio'] = None,
    ) -> 'models.AppBskyFeedPost.CreateRecordResponse':
        """Send post with attached image.

        Note:
            If `profile_identify` is not provided will be sent to the current profile.

        Args:
            text: Text of the post.
            image: Binary image to attach.
            image_alt: Text version of the image.
            profile_identify: Handle or DID. Where to send post.
            reply_to: Root and parent of the post to reply to.
            langs: List of used languages in the post.
            facets: List of facets (rich text items).
            image_aspect_ratio: Aspect ratio of the image.

        Returns:
            :obj:`models.AppBskyFeedPost.CreateRecordResponse`: Reference to the created record.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """
        image_aspect_ratios = None
        if image_aspect_ratio:
            image_aspect_ratios = [image_aspect_ratio]

        return self.send_images(
            text,
            images=[image],
            image_alts=[image_alt],
            profile_identify=profile_identify,
            reply_to=reply_to,
            langs=langs,
            facets=facets,
            image_aspect_ratios=image_aspect_ratios,
        )

    def send_video(
        self,
        text: t.Union[str, TextBuilder],
        video: bytes,
        video_alt: t.Optional[str] = None,
        profile_identify: t.Optional[str] = None,
        reply_to: t.Optional['models.AppBskyFeedPost.ReplyRef'] = None,
        langs: t.Optional[t.List[str]] = None,
        facets: t.Optional[t.List['models.AppBskyRichtextFacet.Main']] = None,
        video_aspect_ratio: t.Optional['models.AppBskyEmbedDefs.AspectRatio'] = None,
    ) -> 'models.AppBskyFeedPost.CreateRecordResponse':
        """Send post with attached video.

        Note:
            If `profile_identify` is not provided will be sent to the current profile.

        Args:
            text: Text of the post.
            video: Binary video to attach.
            video_alt: Text version of the video.
            profile_identify: Handle or DID. Where to send post.
            reply_to: Root and parent of the post to reply to.
            langs: List of used languages in the post.
            facets: List of facets (rich text items).
            video_aspect_ratio: Aspect ratio of the video.

        Returns:
            :obj:`models.AppBskyFeedPost.CreateRecordResponse`: Reference to the created record.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """
        if video_alt is None:
            video_alt = ''

        upload = self.upload_blob(video)

        return self.send_post(
            text,
            profile_identify=profile_identify,
            reply_to=reply_to,
            embed=models.AppBskyEmbedVideo.Main(video=upload.blob, alt=video_alt, aspect_ratio=video_aspect_ratio),
            langs=langs,
            facets=facets,
        )

    def get_post(
        self, post_rkey: str, profile_identify: t.Optional[str] = None, cid: t.Optional[str] = None
    ) -> 'models.AppBskyFeedPost.GetRecordResponse':
        """Get post.

        Args:
            post_rkey: ID (slug) of the post.
            profile_identify: Handler or DID. Who created the post.
            cid: The CID of the version of the post.

        Returns:
            :obj:`models.AppBskyFeedPost.GetRecordResponse`: Post.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """
        repo = self.me and self.me.did
        if profile_identify:
            repo = profile_identify

        if not repo:
            raise LoginRequiredError

        return self.app.bsky.feed.post.get(repo, post_rkey, cid)

    def get_posts(self, uris: t.List[str]) -> 'models.AppBskyFeedGetPosts.Response':
        """Get posts.

        Args:
            uris: Uris (AT URI).

        Example:
            .. code-block:: python

                client.get_posts(['at://did:plc:kvwvcn5iqfooopmyzvb4qzba/app.bsky.feed.post/3k2yihcrp6f2c'])

        Returns:
            :obj:`models.AppBskyFeedGetPosts.Response`: Posts.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """
        return self.app.bsky.feed.get_posts(
            models.AppBskyFeedGetPosts.Params(
                uris=uris,
            )
        )

    def get_post_thread(
        self, uri: str, depth: t.Optional[int] = None, parent_height: t.Optional[int] = None
    ) -> 'models.AppBskyFeedGetPostThread.Response':
        """Get post thread.

        Args:
            uri: AT URI.
            depth: Depth of the thread.
            parent_height: Height of the parent post.

        Returns:
            :obj:`models.AppBskyFeedGetPostThread.Response`: Post thread.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """
        return self.app.bsky.feed.get_post_thread(
            models.AppBskyFeedGetPostThread.Params(
                uri=uri,
                depth=depth,
                parent_height=parent_height,
            )
        )

    def get_likes(
        self, uri: str, cid: t.Optional[str] = None, cursor: t.Optional[str] = None, limit: t.Optional[int] = None
    ) -> 'models.AppBskyFeedGetLikes.Response':
        """Get likes.

        Args:
            uri: AT URI.
            cid: CID.
            cursor: Cursor of the last like in the previous page.
            limit: Limit count of likes to return.

        Returns:
            :obj:`models.AppBskyFeedGetLikes.Response`: Likes.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """
        return self.app.bsky.feed.get_likes(
            models.AppBskyFeedGetLikes.Params(uri=uri, cid=cid, cursor=cursor, limit=limit)
        )

    def get_reposted_by(
        self, uri: str, cid: t.Optional[str] = None, cursor: t.Optional[str] = None, limit: t.Optional[int] = None
    ) -> 'models.AppBskyFeedGetRepostedBy.Response':
        """Get reposted by (reposts).

        Args:
            uri: AT URI.
            cid: CID.
            cursor: Cursor of the last like in the previous page.
            limit: Limit count of likes to return.

        Returns:
            :obj:`models.AppBskyFeedGetRepostedBy.Response`: Reposts.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """
        return self.app.bsky.feed.get_reposted_by(
            models.AppBskyFeedGetRepostedBy.Params(uri=uri, cid=cid, cursor=cursor, limit=limit)
        )

    def get_timeline(
        self, algorithm: t.Optional[str] = None, cursor: t.Optional[str] = None, limit: t.Optional[int] = None
    ) -> 'models.AppBskyFeedGetTimeline.Response':
        """Get home timeline.

        Args:
            algorithm: Algorithm.
            cursor: Cursor of the last like in the previous page.
            limit: Limit count of likes to return.

        Returns:
            :obj:`models.AppBskyFeedGetTimeline.Response`: Home timeline.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """
        return self.app.bsky.feed.get_timeline(
            models.AppBskyFeedGetTimeline.Params(algorithm=algorithm, cursor=cursor, limit=limit)
        )

    def get_author_feed(
        self,
        actor: str,
        cursor: t.Optional[str] = None,
        filter: t.Optional[str] = None,
        limit: t.Optional[int] = None,
        include_pins: bool = False,
    ) -> 'models.AppBskyFeedGetAuthorFeed.Response':
        """Get author (profile) feed.

        Args:
            actor: Actor (handle or DID).
            cursor: Cursor of the last like in the previous page.
            filter: Filter.
            limit: Limit count of likes to return.
            include_pins: Include pins.

        Returns:
            :obj:`models.AppBskyFeedGetAuthorFeed.Response`: Feed.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """
        return self.app.bsky.feed.get_author_feed(
            models.AppBskyFeedGetAuthorFeed.Params(
                actor=actor, cursor=cursor, filter=filter, include_pins=include_pins, limit=limit
            )
        )

    def like(self, uri: str, cid: str) -> 'models.AppBskyFeedLike.CreateRecordResponse':
        """Like the record.

        Args:
            cid: The CID of the record.
            uri: The URI of the record.

        Note:
            Record could be post, custom feed, etc.

        Returns:
            :obj:`models.AppBskyFeedLike.CreateRecordResponse`: Reference to the created record.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """
        subject_obj = models.ComAtprotoRepoStrongRef.Main(cid=cid, uri=uri)

        repo = self.me and self.me.did
        if not repo:
            raise LoginRequiredError

        record = models.AppBskyFeedLike.Record(created_at=self.get_current_time_iso(), subject=subject_obj)
        return self.app.bsky.feed.like.create(repo, record)

    def unlike(self, like_uri: str) -> bool:
        """Unlike the post.

        Args:
            like_uri: AT URI of the like.

        Returns:
            :obj:`bool`: Success status.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """
        uri = AtUri.from_str(like_uri)
        return self.app.bsky.feed.like.delete(uri.hostname, uri.rkey)

    def repost(self, uri: str, cid: str) -> 'models.AppBskyFeedRepost.CreateRecordResponse':
        """Repost post.

        Args:
            cid: The CID of the post.
            uri: The URI of the post.

        Returns:
            :obj:`models.AppBskyFeedRepost.CreateRecordResponse`: Reference to the reposted record.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """
        subject_obj = models.ComAtprotoRepoStrongRef.Main(cid=cid, uri=uri)

        repo = self.me and self.me.did
        if not repo:
            raise LoginRequiredError

        record = models.AppBskyFeedRepost.Record(created_at=self.get_current_time_iso(), subject=subject_obj)
        return self.app.bsky.feed.repost.create(repo, record)

    def unrepost(self, repost_uri: str) -> bool:
        """Unrepost the post (delete repost).

        Args:
            repost_uri: AT URI of the repost.

        Returns:
            :obj:`bool`: Success status.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """
        uri = AtUri.from_str(repost_uri)
        return self.app.bsky.feed.repost.delete(uri.hostname, uri.rkey)

    def follow(self, subject: str) -> 'models.AppBskyGraphFollow.CreateRecordResponse':
        """Follow the profile.

        Args:
            subject: DID of the profile.

        Returns:
            :obj:`models.AppBskyGraphFollow.CreateRecordResponse`: Reference to the created record.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """
        repo = self.me and self.me.did
        if not repo:
            raise LoginRequiredError

        record = models.AppBskyGraphFollow.Record(created_at=self.get_current_time_iso(), subject=subject)
        return self.app.bsky.graph.follow.create(repo, record)

    def unfollow(self, follow_uri: str) -> bool:
        """Unfollow the profile.

        Args:
            follow_uri: AT URI of the follow.

        Returns:
            :obj:`bool`: Success status.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """
        uri = AtUri.from_str(follow_uri)
        return self.app.bsky.graph.follow.delete(uri.hostname, uri.rkey)

    def get_follows(
        self, actor: str, cursor: t.Optional[str] = None, limit: t.Optional[int] = None
    ) -> 'models.AppBskyGraphGetFollows.Response':
        """Get follows of the profile.

        Args:
            actor: Actor (handle or DID).
            cursor: Cursor of the next page.
            limit: Limit count of follows to return.

        Returns:
            :obj:`models.AppBskyGraphGetFollows.Response`: Follows.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """
        return self.app.bsky.graph.get_follows(
            models.AppBskyGraphGetFollows.Params(actor=actor, cursor=cursor, limit=limit)
        )

    def get_followers(
        self, actor: str, cursor: t.Optional[str] = None, limit: t.Optional[int] = None
    ) -> 'models.AppBskyGraphGetFollowers.Response':
        """Get followers of the profile.

        Args:
            actor: Actor (handle or DID).
            cursor: Cursor of the next page.
            limit: Limit count of followers to return.

        Returns:
            :obj:`models.AppBskyGraphGetFollowers.Response`: Followers.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """
        return self.app.bsky.graph.get_followers(
            models.AppBskyGraphGetFollowers.Params(actor=actor, cursor=cursor, limit=limit)
        )

    def get_profile(self, actor: str) -> 'models.AppBskyActorDefs.ProfileViewDetailed':
        """Get profile.

        Args:
            actor: Actor (handle or DID).

        Returns:
            :obj:`models.AppBskyActorDefs.ProfileViewDetailed`: Profile.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """
        return self.app.bsky.actor.get_profile(models.AppBskyActorGetProfile.Params(actor=actor))

    def get_profiles(self, actors: t.List[str]) -> 'models.AppBskyActorGetProfiles.Response':
        """Get profiles.

        Args:
            actors: List of actors (handles or DIDs).

        Returns:
            :obj:`models.AppBskyActorGetProfiles.Response`: Profiles.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """
        return self.app.bsky.actor.get_profiles(models.AppBskyActorGetProfiles.Params(actors=actors))

    def mute(self, actor: str) -> bool:
        """Mute actor (profile).

        Args:
            actor: Actor (handle or DID).

        Returns:
            :obj:`bool`: Success status.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """
        return self.app.bsky.graph.mute_actor(models.AppBskyGraphMuteActor.Data(actor=actor))

    def unmute(self, actor: str) -> bool:
        """Unmute actor (profile).

        Args:
            actor: Actor (handle or DID).

        Returns:
            :obj:`bool`: Success status.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """
        return self.app.bsky.graph.unmute_actor(models.AppBskyGraphUnmuteActor.Data(actor=actor))

    def resolve_handle(self, handle: str) -> 'models.ComAtprotoIdentityResolveHandle.Response':
        """Resolve the handle.

        Args:
            handle: Handle.

        Returns:
            :obj:`models.ComAtprotoIdentityResolveHandle.Response`: Resolved handle (DID).

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """
        return self.com.atproto.identity.resolve_handle(models.ComAtprotoIdentityResolveHandle.Params(handle=handle))

    def update_handle(self, handle: str) -> bool:
        """Update the handle.

        Args:
            handle: New handle.

        Returns:
            :obj:`bool`: Success status.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """
        return self.com.atproto.identity.update_handle(models.ComAtprotoIdentityUpdateHandle.Data(handle=handle))

    def upload_blob(self, data: bytes) -> 'models.ComAtprotoRepoUploadBlob.Response':
        """Upload blob.

        Args:
            data: Binary data.

        Returns:
            :obj:`models.ComAtprotoRepoUploadBlob.Response`: Uploaded blob reference.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """
        return self.com.atproto.repo.upload_blob(data)

    #: Alias for :attr:`unfollow`
    delete_follow = unfollow
    #: Alias for :attr:`unlike`
    delete_like = unlike
    #: Alias for :attr:`unrepost`
    delete_repost = unrepost
    #: Alias for :attr:`send_post`
    post = send_post
    #: Alias for :attr:`delete_post`
    unsend = delete_post



"""AT Protocol string format validation."""

import re
from datetime import datetime
from functools import wraps
from typing import Callable, Mapping, Set, Union, cast
from urllib.parse import urlparse

from atproto_core.exceptions import InvalidNsidError
from atproto_core.nsid import validate_nsid as atproto_core_validate_nsid
from pydantic import BeforeValidator, Field, ValidationInfo
from pydantic_core import core_schema
from typing_extensions import Annotated, Literal

_OPT_IN_KEY: Literal['strict_string_format'] = 'strict_string_format'

# constants
MAX_HANDLE_LENGTH: int = 253
MAX_RECORD_KEY_LENGTH: int = 512
MAX_URI_LENGTH: int = 8 * 1024
MIN_CID_LENGTH: int = 8
TID_LENGTH: int = 13
INVALID_RECORD_KEYS: Set[str] = {'.', '..'}
MAX_DID_LENGTH: int = 2048  # Method-specific identifier max length
MAX_AT_URI_LENGTH: int = 8 * 1024

# patterns
DOMAIN_RE = re.compile(
    r'^'
    r'([a-zA-Z0-9]([a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?\.)'  # First/middle segments
    r'+'  # One or more of those segments
    r'[a-zA-Z]([a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?'  # TLD: must start with letter, then optional chars
    r'$'
)
DID_RE = re.compile(
    r'^did:'  # Required prefix
    r'[a-z]+:'  # method-name (lowercase only)
    r'[a-zA-Z0-9._:%-]*'  # method-specific-id with allowed chars
    r'[a-zA-Z0-9._-]$'  # must end with allowed char (not : or %)
)
LANG_RE = re.compile(r'^(i|[a-z]{2,3})(-[A-Za-z0-9-]+)?$')
RKEY_RE = re.compile(r'^[A-Za-z0-9._:~-]{1,512}$')
TID_RE = re.compile(rf'^[2-7a-z]{{{TID_LENGTH}}}$')
CID_RE = re.compile(r'^[A-Za-z0-9+]{8,}$')
AT_URI_RE = re.compile(
    r'^at://'  # Must start with at://
    r'('  # Authority group start
    # For DIDs: Only allowed chars are letters, numbers, period, hyphen, and percent
    r'did:[a-z]+:[a-zA-Z0-9.-]+'  # Notice removed underscore from allowed chars
    r'|'  # or
    # Handle: require 2+ segments, TLD can't start with digit
    r'[a-zA-Z0-9]([a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?'  # First segment
    r'(\.[a-zA-Z0-9]([a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?)*'  # Middle segments
    r'\.[a-zA-Z][a-zA-Z0-9-]*'  # TLD must start with letter
    r')'  # Authority group end
    r'(?:'  # Optional path group
    # NSID: Match core validator rules - last segment must start with letter then allow letters/numbers/hyphens
    r'/[a-z][a-zA-Z0-9-]*(?:\.[a-z][a-zA-Z0-9-]*)*\.[a-zA-Z][a-zA-Z0-9-]*'  # NSID
    r'(?:/(?!\.\.?(?:/|$))[A-Za-z0-9._:~-]+)?'  # Record key: must not be . or .. and must not be empty
    r')?$'
)


class _NamedValidator:
    """Decorator to add a __str__ attribute to a validation function."""

    def __init__(self, validate_fn: Callable[..., str]) -> None:
        self.validate_fn = validate_fn

    def __call__(self, v: str, info: ValidationInfo) -> str:
        return self.validate_fn(v, info)

    def __str__(self) -> str:
        func_str = f':func:`string_formats.{self.validate_fn.__name__}`'
        return f'Validated by: {func_str} (only when `strict_string_format=True`)'


def only_validate_if_strict(validate_fn: Callable[..., str]) -> Callable[..., str]:
    """Skip pydantic validation if not opting into strict validation via context."""

    @wraps(validate_fn)
    def wrapper(v: str, info: ValidationInfo) -> str:
        """Could likely be generalized to support arbitrary signatures."""
        if info and isinstance(info.context, Mapping) and info.context.get(_OPT_IN_KEY, False):
            return cast(core_schema.WithInfoValidatorFunction, validate_fn)(v, info)
        return v

    return wrapper


@only_validate_if_strict
def validate_handle(v: str, _: ValidationInfo) -> str:
    """Validate an AT Protocol Handle Identifier.

    A handle must be a valid domain name with:

    - 2+ segments separated by dots

    - ASCII alphanumeric characters and hyphens only

    - 1-63 chars per segment

    - Max 253 chars total

    - Last segment cannot start with a digit

    Args:
        v: The handle to validate (e.g. alice.bsky.social)

    Returns:
        The validated handle

    Raises:
        ValueError: If handle format is invalid
    """
    # Check ASCII first
    if not v.isascii():
        raise ValueError('Invalid handle: must contain only ASCII characters')

    # Use the spec's reference regex
    if not DOMAIN_RE.match(v.lower()) or len(v) > MAX_HANDLE_LENGTH:
        raise ValueError(
            f'Invalid handle: must be a domain name (e.g. user.bsky.social) with max length {MAX_HANDLE_LENGTH}'
        )

    return v


@only_validate_if_strict
def validate_did(v: str, _: ValidationInfo) -> str:
    """Validate a Decentralized Identifiers (DID).

    A DID must follow the pattern:

    - Format: did:method:identifier

    - Method must be lowercase letters

    - Identifier allows alphanumeric chars, dots, underscores, hyphens, and percent

    - Max 2KB length

    - No /?#[]@ characters allowed

    - Valid percent-encoding if used

    Args:
        v: The DID to validate (e.g. did:plc:z72i7hdynmk6r22z27h6tvur)

    Returns:
        The validated DID

    Raises:
        ValueError: If DID format is invalid
    """
    # Check length first
    if len(v) > MAX_DID_LENGTH:
        raise ValueError(f'Invalid DID: must be under {MAX_DID_LENGTH} chars')

    # Check for invalid characters
    if any(c in v for c in '/?#[]@'):
        raise ValueError('Invalid DID: cannot contain /, ?, #, [, ], or @ characters')

    # Check for invalid percent encoding
    if '%' in v:
        percent_segments = v.split('%')[1:]
        for segment in percent_segments:
            if len(segment) < 2 or not segment[:2].isalnum():
                raise ValueError('Invalid DID: invalid percent-encoding')

    # Check against regex pattern
    if not DID_RE.match(v):
        raise ValueError('Invalid DID: must be in format did:method:identifier (e.g. did:plc:1234abcd)')

    return v


@only_validate_if_strict
def validate_nsid(v: str, _: ValidationInfo) -> str:
    """Validate an AT Protocol NSID (Namespaced Identifier).

    An NSID must have:

    - 3+ segments separated by dots

    - Reversed domain name (lowercase alphanumeric + hyphen)

    - Name segment (letters only)

    - Max 317 chars total

    - No segments ending in numbers except last segment

    - No @_*#! special characters

    - Max 63 chars per segment

    - Non-leading digits are allowed in the name (last) segment

    Args:
        v: The NSID to validate (e.g. app.bsky.feed.post)

    Returns:
        The validated NSID

    Raises:
        ValueError: If NSID format is invalid
    """
    try:
        atproto_core_validate_nsid(v)
        return v
    except InvalidNsidError as e:
        raise ValueError(str(e)) from None


@only_validate_if_strict
def validate_language(v: str, _: ValidationInfo) -> str:
    """Validate an ISO language code.

    Must match pattern:

    - 2-3 letter language code or 'i'

    - Optional subtag with alphanumeric chars and hyphens

    Args:
        v: The language code to validate (e.g. en or en-US)

    Returns:
        The validated language code

    Raises:
        ValueError: If language code format is invalid
    """
    if not LANG_RE.match(v):
        raise ValueError('Invalid language code: must be ISO language code (e.g. en or en-US)')
    return v


@only_validate_if_strict
def validate_record_key(v: str, _: ValidationInfo) -> str:
    """Validate an AT Protocol Record Key (rkey).

    A record key must:

    - Be 1-512 characters

    - Contain only alphanumeric chars, dots, underscores, colons, tildes, or hyphens

    - Not be "." or ".."

    Args:
        v: The record key to validate (e.g. 3jxtb5w2hkt2m)

    Returns:
        The validated record key

    Raises:
        ValueError: If record key format is invalid
    """
    if v in INVALID_RECORD_KEYS or not RKEY_RE.match(v):
        raise ValueError(
            'Invalid record key: must contain only alphanumeric, dot, underscore, colon, tilde, or hyphen characters'
        )
    return v


@only_validate_if_strict
def validate_cid(v: str, _: ValidationInfo) -> str:
    """Validate a Content Identifier (CID).

    Must be:

    - Minimum 8 characters

    - Alphanumeric characters and plus signs only

    Args:
        v: The CID to validate (e.g. bafybeigdyrzt5sfp7udm7hu76uh7y26nf3efuylqabf3oclgtqy55fbzdi)

    Returns:
        The validated CID

    Raises:
        ValueError: If CID format is invalid
    """
    if not CID_RE.match(v):
        raise ValueError('Invalid CID: must be a valid Content Identifier with minimum length 8')
    return v


@only_validate_if_strict
def validate_at_uri(v: str, _: ValidationInfo) -> str:
    """Validate an AT Protocol URI.

    Must follow pattern:

    - Starts with at://

    - Contains handle or DID

    - Optional /collection/record-key path

    - Max 8KB length

    - No query parameters or fragments

    Args:
        v: The AT-URI to validate (e.g. at://alice.bsky.social/app.bsky.feed.post/3jxtb5w2hkt2m)

    Returns:
        The validated AT-URI

    Raises:
        ValueError: If AT-URI format is invalid
    """
    if len(v) >= MAX_AT_URI_LENGTH:
        raise ValueError(f'Invalid AT-URI: must be under {MAX_AT_URI_LENGTH} chars')

    if not AT_URI_RE.match(v):
        raise ValueError('Invalid AT-URI: invalid format')

    return v


@only_validate_if_strict
def validate_datetime(v: str, _: ValidationInfo) -> str:  # noqa: C901
    """Validate an ISO 8601/RFC 3339 datetime string.

    Requirements:

    - Must use uppercase T as time separator

    - Must include seconds (HH:MM:SS)

    - Must have timezone (Z or ±HH:MM)

    - No -00:00 timezone allowed

    - Valid fractional seconds format if used (any precision)

    - No whitespace allowed

    - Years can have leading zeros

    Args:
        v: The datetime string to validate (e.g. 2024-11-24T06:02:00Z)

    Returns:
        The validated datetime string

    Raises:
        ValueError: If datetime format is invalid
    """
    # Must not have whitespace
    if v != v.strip():
        raise ValueError('Invalid datetime: no whitespace allowed')

    # Must contain uppercase T
    if 'T' not in v:
        raise ValueError('Invalid datetime: must contain uppercase T separator')

    # Split into date and time parts
    try:
        time_str = v.split('T')[1]
    except IndexError:
        raise ValueError('Invalid datetime: invalid format') from None

    # Extract the time part before any timezone
    if 'Z' in time_str:
        time_part = time_str.split('Z')[0]
    elif '+' in time_str:
        time_part = time_str.split('+')[0]
    elif '-' in time_str:
        # Find the last '-' in case year is negative
        time_part = time_str.rsplit('-', 1)[0]
    else:
        time_part = time_str

    # Check HH:MM:SS format
    time_segments = time_part.split(':')
    if len(time_segments) != 3:
        raise ValueError('Invalid datetime: seconds are required')

    # If has decimal point, must have digits after it
    if '.' in time_segments[2] and not re.search(r'\.\d+', time_segments[2]):
        raise ValueError('Invalid datetime: invalid fractional seconds format')

    # Must have valid timezone
    if v.endswith('-00:00'):
        raise ValueError('Invalid datetime: -00:00 timezone not allowed')

    # Must end with Z or valid timezone offset
    if not (v.endswith('Z') or re.search(r'[+-]\d{2}:\d{2}$', v)):
        raise ValueError('Invalid datetime: must include timezone')

    # Final validation using datetime.fromisoformat
    try:
        # Handle both Z and explicit timezone formats
        datetime_str = v
        if v.endswith('Z'):
            # For Python 3.10 compatibility, normalize fractional seconds to 6 digits
            # see https://docs.python.org/3/whatsnew/3.11.html#datetime and https://github.com/python/cpython/issues/80010
            if '.' in v:
                base, frac = v[:-1].split('.')
                # Pad or truncate to exactly 6 digits
                frac = (frac + '000000')[:6]
                datetime_str = f'{base}.{frac}+00:00'
            else:
                datetime_str = v[:-1] + '+00:00'
        datetime.fromisoformat(datetime_str)
        return v
    except ValueError as e:
        raise ValueError(f'Invalid datetime: {e!s}') from None


@only_validate_if_strict
def validate_tid(v: str, _: ValidationInfo) -> str:
    """Validate an AT Protocol TID (Timestamp Identifiers).

    Must be:

    - Exactly 13 characters

    - Only lowercase letters and numbers 2-7

    - First byte's high bit (0x40) must be 0

    Args:
        v: The TID to validate (e.g. 3jxtb5w2hkt2m)

    Returns:
        The validated TID

    Raises:
        ValueError: If TID format is invalid
    """
    if not TID_RE.match(v) or (ord(v[0]) & 0x40):
        raise ValueError(f'Invalid TID: must be exactly {TID_LENGTH} lowercase letters/numbers')
    return v


@only_validate_if_strict
def validate_uri(v: str, _: ValidationInfo) -> str:
    """Validate a standard URI.

    Requirements:

    - Must have a scheme starting with a letter

    - Must have authority (netloc) or path/query/fragment

    - Max 8KB length

    - No spaces allowed

    - Must follow RFC-3986 format

    Args:
        v: The URI to validate (e.g. https://example.com/path)

    Returns:
        The validated URI

    Raises:
        ValueError: If URI format is invalid
    """
    if len(v) >= MAX_URI_LENGTH or ' ' in v:
        raise ValueError(f'Invalid URI: must be under {MAX_URI_LENGTH} chars and not contain spaces')
    parsed = urlparse(v)
    if not (
        parsed.scheme
        and parsed.scheme[0].isalpha()
        and (parsed.netloc or parsed.path or parsed.query or parsed.fragment)
    ):
        raise ValueError('Invalid URI: must be a valid URI with scheme and authority/path (e.g. https://example.com)')
    return v


Handle = Annotated[str, BeforeValidator(_NamedValidator(validate_handle))]
Did = Annotated[str, BeforeValidator(_NamedValidator(validate_did))]
Nsid = Annotated[str, BeforeValidator(_NamedValidator(validate_nsid))]
Language = Annotated[str, BeforeValidator(_NamedValidator(validate_language))]
RecordKey = Annotated[str, BeforeValidator(_NamedValidator(validate_record_key))]
Cid = Annotated[str, BeforeValidator(_NamedValidator(validate_cid))]
AtUri = Annotated[str, BeforeValidator(_NamedValidator(validate_at_uri))]
DateTime = Annotated[
    str, BeforeValidator(_NamedValidator(validate_datetime))
]  # see https://github.com/python-pendulum/pendulum/issues/844
Tid = Annotated[str, BeforeValidator(_NamedValidator(validate_tid))]
Uri = Annotated[str, BeforeValidator(_NamedValidator(validate_uri))]

AtIdentifier = Union[Handle, Did]

# Any valid ATProto string format
AtProtoString = Annotated[
    Union[Handle, Did, Nsid, AtUri, Cid, DateTime, Tid, RecordKey, Uri, Language],
    Field(description='ATProto string format'),
]



from atproto_core.exceptions import AtProtocolError


class DidPlcResolverError(AtProtocolError): ...


class DidWebResolverError(AtProtocolError): ...


class PoorlyFormattedDidError(AtProtocolError): ...


class UnsupportedDidWebPathError(AtProtocolError): ...


class UnsupportedDidMethodError(AtProtocolError): ...


class PoorlyFormattedDidDocumentError(AtProtocolError): ...


class DidNotFoundError(AtProtocolError): ...


class AtprotoDataParseError(AtProtocolError): ...



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2024 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t

import typing_extensions as te
from pydantic import Field

from atproto_client.models import string_formats

if t.TYPE_CHECKING:
    from atproto_client import models
    from atproto_client.models.unknown_type import UnknownType
from atproto_client.models import base


class PostView(base.ModelBase):
    """Definition model for :obj:`app.bsky.feed.defs`."""

    author: 'models.AppBskyActorDefs.ProfileViewBasic'  #: Author.
    cid: string_formats.Cid  #: Cid.
    indexed_at: string_formats.DateTime  #: Indexed at.
    record: 'UnknownType'  #: Record.
    uri: string_formats.AtUri  #: Uri.
    embed: t.Optional[
        te.Annotated[
            t.Union[
                'models.AppBskyEmbedImages.View',
                'models.AppBskyEmbedVideo.View',
                'models.AppBskyEmbedExternal.View',
                'models.AppBskyEmbedRecord.View',
                'models.AppBskyEmbedRecordWithMedia.View',
            ],
            Field(default=None, discriminator='py_type'),
        ]
    ] = None  #: Embed.
    labels: t.Optional[t.List['models.ComAtprotoLabelDefs.Label']] = None  #: Labels.
    like_count: t.Optional[int] = None  #: Like count.
    quote_count: t.Optional[int] = None  #: Quote count.
    reply_count: t.Optional[int] = None  #: Reply count.
    repost_count: t.Optional[int] = None  #: Repost count.
    threadgate: t.Optional['models.AppBskyFeedDefs.ThreadgateView'] = None  #: Threadgate.
    viewer: t.Optional['models.AppBskyFeedDefs.ViewerState'] = None  #: Viewer.

    py_type: t.Literal['app.bsky.feed.defs#postView'] = Field(
        default='app.bsky.feed.defs#postView', alias='$type', frozen=True
    )


class ViewerState(base.ModelBase):
    """Definition model for :obj:`app.bsky.feed.defs`. Metadata about the requesting account's relationship with the subject content. Only has meaningful content for authed requests."""

    embedding_disabled: t.Optional[bool] = None  #: Embedding disabled.
    like: t.Optional[string_formats.AtUri] = None  #: Like.
    pinned: t.Optional[bool] = None  #: Pinned.
    reply_disabled: t.Optional[bool] = None  #: Reply disabled.
    repost: t.Optional[string_formats.AtUri] = None  #: Repost.
    thread_muted: t.Optional[bool] = None  #: Thread muted.

    py_type: t.Literal['app.bsky.feed.defs#viewerState'] = Field(
        default='app.bsky.feed.defs#viewerState', alias='$type', frozen=True
    )


class ThreadContext(base.ModelBase):
    """Definition model for :obj:`app.bsky.feed.defs`. Metadata about this post within the context of the thread it is in."""

    root_author_like: t.Optional[string_formats.AtUri] = None  #: Root author like.

    py_type: t.Literal['app.bsky.feed.defs#threadContext'] = Field(
        default='app.bsky.feed.defs#threadContext', alias='$type', frozen=True
    )


class FeedViewPost(base.ModelBase):
    """Definition model for :obj:`app.bsky.feed.defs`."""

    post: 'models.AppBskyFeedDefs.PostView'  #: Post.
    feed_context: t.Optional[str] = Field(
        default=None, max_length=2000
    )  #: Context provided by feed generator that may be passed back alongside interactions.
    reason: t.Optional[
        te.Annotated[
            t.Union['models.AppBskyFeedDefs.ReasonRepost', 'models.AppBskyFeedDefs.ReasonPin'],
            Field(default=None, discriminator='py_type'),
        ]
    ] = None  #: Reason.
    reply: t.Optional['models.AppBskyFeedDefs.ReplyRef'] = None  #: Reply.

    py_type: t.Literal['app.bsky.feed.defs#feedViewPost'] = Field(
        default='app.bsky.feed.defs#feedViewPost', alias='$type', frozen=True
    )


class ReplyRef(base.ModelBase):
    """Definition model for :obj:`app.bsky.feed.defs`."""

    parent: te.Annotated[
        t.Union[
            'models.AppBskyFeedDefs.PostView',
            'models.AppBskyFeedDefs.NotFoundPost',
            'models.AppBskyFeedDefs.BlockedPost',
        ],
        Field(discriminator='py_type'),
    ]  #: Parent.
    root: te.Annotated[
        t.Union[
            'models.AppBskyFeedDefs.PostView',
            'models.AppBskyFeedDefs.NotFoundPost',
            'models.AppBskyFeedDefs.BlockedPost',
        ],
        Field(discriminator='py_type'),
    ]  #: Root.
    grandparent_author: t.Optional['models.AppBskyActorDefs.ProfileViewBasic'] = (
        None  #: When parent is a reply to another post, this is the author of that post.
    )

    py_type: t.Literal['app.bsky.feed.defs#replyRef'] = Field(
        default='app.bsky.feed.defs#replyRef', alias='$type', frozen=True
    )


class ReasonRepost(base.ModelBase):
    """Definition model for :obj:`app.bsky.feed.defs`."""

    by: 'models.AppBskyActorDefs.ProfileViewBasic'  #: By.
    indexed_at: string_formats.DateTime  #: Indexed at.

    py_type: t.Literal['app.bsky.feed.defs#reasonRepost'] = Field(
        default='app.bsky.feed.defs#reasonRepost', alias='$type', frozen=True
    )


class ReasonPin(base.ModelBase):
    """Definition model for :obj:`app.bsky.feed.defs`."""

    py_type: t.Literal['app.bsky.feed.defs#reasonPin'] = Field(
        default='app.bsky.feed.defs#reasonPin', alias='$type', frozen=True
    )


class ThreadViewPost(base.ModelBase):
    """Definition model for :obj:`app.bsky.feed.defs`."""

    post: 'models.AppBskyFeedDefs.PostView'  #: Post.
    parent: t.Optional[
        te.Annotated[
            t.Union[
                'models.AppBskyFeedDefs.ThreadViewPost',
                'models.AppBskyFeedDefs.NotFoundPost',
                'models.AppBskyFeedDefs.BlockedPost',
            ],
            Field(default=None, discriminator='py_type'),
        ]
    ] = None  #: Parent.
    replies: t.Optional[
        t.List[
            te.Annotated[
                t.Union[
                    'models.AppBskyFeedDefs.ThreadViewPost',
                    'models.AppBskyFeedDefs.NotFoundPost',
                    'models.AppBskyFeedDefs.BlockedPost',
                ],
                Field(discriminator='py_type'),
            ]
        ]
    ] = None  #: Replies.
    thread_context: t.Optional['models.AppBskyFeedDefs.ThreadContext'] = None  #: Thread context.

    py_type: t.Literal['app.bsky.feed.defs#threadViewPost'] = Field(
        default='app.bsky.feed.defs#threadViewPost', alias='$type', frozen=True
    )


class NotFoundPost(base.ModelBase):
    """Definition model for :obj:`app.bsky.feed.defs`."""

    not_found: bool = Field(frozen=True)  #: Not found.
    uri: string_formats.AtUri  #: Uri.

    py_type: t.Literal['app.bsky.feed.defs#notFoundPost'] = Field(
        default='app.bsky.feed.defs#notFoundPost', alias='$type', frozen=True
    )


class BlockedPost(base.ModelBase):
    """Definition model for :obj:`app.bsky.feed.defs`."""

    author: 'models.AppBskyFeedDefs.BlockedAuthor'  #: Author.
    blocked: bool = Field(frozen=True)  #: Blocked.
    uri: string_formats.AtUri  #: Uri.

    py_type: t.Literal['app.bsky.feed.defs#blockedPost'] = Field(
        default='app.bsky.feed.defs#blockedPost', alias='$type', frozen=True
    )


class BlockedAuthor(base.ModelBase):
    """Definition model for :obj:`app.bsky.feed.defs`."""

    did: string_formats.Did  #: Did.
    viewer: t.Optional['models.AppBskyActorDefs.ViewerState'] = None  #: Viewer.

    py_type: t.Literal['app.bsky.feed.defs#blockedAuthor'] = Field(
        default='app.bsky.feed.defs#blockedAuthor', alias='$type', frozen=True
    )


class GeneratorView(base.ModelBase):
    """Definition model for :obj:`app.bsky.feed.defs`."""

    cid: string_formats.Cid  #: Cid.
    creator: 'models.AppBskyActorDefs.ProfileView'  #: Creator.
    did: string_formats.Did  #: Did.
    display_name: str  #: Display name.
    indexed_at: string_formats.DateTime  #: Indexed at.
    uri: string_formats.AtUri  #: Uri.
    accepts_interactions: t.Optional[bool] = None  #: Accepts interactions.
    avatar: t.Optional[string_formats.Uri] = None  #: Avatar.
    content_mode: t.Optional[
        t.Union['models.AppBskyFeedDefs.ContentModeUnspecified', 'models.AppBskyFeedDefs.ContentModeVideo', str]
    ] = None  #: Content mode.
    description: t.Optional[str] = Field(default=None, max_length=3000)  #: Description.
    description_facets: t.Optional[t.List['models.AppBskyRichtextFacet.Main']] = None  #: Description facets.
    labels: t.Optional[t.List['models.ComAtprotoLabelDefs.Label']] = None  #: Labels.
    like_count: t.Optional[int] = Field(default=None, ge=0)  #: Like count.
    viewer: t.Optional['models.AppBskyFeedDefs.GeneratorViewerState'] = None  #: Viewer.

    py_type: t.Literal['app.bsky.feed.defs#generatorView'] = Field(
        default='app.bsky.feed.defs#generatorView', alias='$type', frozen=True
    )


class GeneratorViewerState(base.ModelBase):
    """Definition model for :obj:`app.bsky.feed.defs`."""

    like: t.Optional[string_formats.AtUri] = None  #: Like.

    py_type: t.Literal['app.bsky.feed.defs#generatorViewerState'] = Field(
        default='app.bsky.feed.defs#generatorViewerState', alias='$type', frozen=True
    )


class SkeletonFeedPost(base.ModelBase):
    """Definition model for :obj:`app.bsky.feed.defs`."""

    post: string_formats.AtUri  #: Post.
    feed_context: t.Optional[str] = Field(
        default=None, max_length=2000
    )  #: Context that will be passed through to client and may be passed to feed generator back alongside interactions.
    reason: t.Optional[
        te.Annotated[
            t.Union['models.AppBskyFeedDefs.SkeletonReasonRepost', 'models.AppBskyFeedDefs.SkeletonReasonPin'],
            Field(default=None, discriminator='py_type'),
        ]
    ] = None  #: Reason.

    py_type: t.Literal['app.bsky.feed.defs#skeletonFeedPost'] = Field(
        default='app.bsky.feed.defs#skeletonFeedPost', alias='$type', frozen=True
    )


class SkeletonReasonRepost(base.ModelBase):
    """Definition model for :obj:`app.bsky.feed.defs`."""

    repost: string_formats.AtUri  #: Repost.

    py_type: t.Literal['app.bsky.feed.defs#skeletonReasonRepost'] = Field(
        default='app.bsky.feed.defs#skeletonReasonRepost', alias='$type', frozen=True
    )


class SkeletonReasonPin(base.ModelBase):
    """Definition model for :obj:`app.bsky.feed.defs`."""

    py_type: t.Literal['app.bsky.feed.defs#skeletonReasonPin'] = Field(
        default='app.bsky.feed.defs#skeletonReasonPin', alias='$type', frozen=True
    )


class ThreadgateView(base.ModelBase):
    """Definition model for :obj:`app.bsky.feed.defs`."""

    cid: t.Optional[string_formats.Cid] = None  #: Cid.
    lists: t.Optional[t.List['models.AppBskyGraphDefs.ListViewBasic']] = None  #: Lists.
    record: t.Optional['UnknownType'] = None  #: Record.
    uri: t.Optional[string_formats.AtUri] = None  #: Uri.

    py_type: t.Literal['app.bsky.feed.defs#threadgateView'] = Field(
        default='app.bsky.feed.defs#threadgateView', alias='$type', frozen=True
    )


class Interaction(base.ModelBase):
    """Definition model for :obj:`app.bsky.feed.defs`."""

    event: t.Optional[
        t.Union[
            'models.AppBskyFeedDefs.RequestLess',
            'models.AppBskyFeedDefs.RequestMore',
            'models.AppBskyFeedDefs.ClickthroughItem',
            'models.AppBskyFeedDefs.ClickthroughAuthor',
            'models.AppBskyFeedDefs.ClickthroughReposter',
            'models.AppBskyFeedDefs.ClickthroughEmbed',
            'models.AppBskyFeedDefs.InteractionSeen',
            'models.AppBskyFeedDefs.InteractionLike',
            'models.AppBskyFeedDefs.InteractionRepost',
            'models.AppBskyFeedDefs.InteractionReply',
            'models.AppBskyFeedDefs.InteractionQuote',
            'models.AppBskyFeedDefs.InteractionShare',
            str,
        ]
    ] = None  #: Event.
    feed_context: t.Optional[str] = Field(
        default=None, max_length=2000
    )  #: Context on a feed item that was originally supplied by the feed generator on getFeedSkeleton.
    item: t.Optional[string_formats.AtUri] = None  #: Item.

    py_type: t.Literal['app.bsky.feed.defs#interaction'] = Field(
        default='app.bsky.feed.defs#interaction', alias='$type', frozen=True
    )


RequestLess = t.Literal[
    'app.bsky.feed.defs#requestLess'
]  #: Request that less content like the given feed item be shown in the feed

RequestMore = t.Literal[
    'app.bsky.feed.defs#requestMore'
]  #: Request that more content like the given feed item be shown in the feed

ClickthroughItem = t.Literal['app.bsky.feed.defs#clickthroughItem']  #: User clicked through to the feed item

ClickthroughAuthor = t.Literal[
    'app.bsky.feed.defs#clickthroughAuthor'
]  #: User clicked through to the author of the feed item

ClickthroughReposter = t.Literal[
    'app.bsky.feed.defs#clickthroughReposter'
]  #: User clicked through to the reposter of the feed item

ClickthroughEmbed = t.Literal[
    'app.bsky.feed.defs#clickthroughEmbed'
]  #: User clicked through to the embedded content of the feed item

ContentModeUnspecified = t.Literal[
    'app.bsky.feed.defs#contentModeUnspecified'
]  #: Declares the feed generator returns any types of posts.

ContentModeVideo = t.Literal[
    'app.bsky.feed.defs#contentModeVideo'
]  #: Declares the feed generator returns posts containing app.bsky.embed.video embeds.

InteractionSeen = t.Literal['app.bsky.feed.defs#interactionSeen']  #: Feed item was seen by user

InteractionLike = t.Literal['app.bsky.feed.defs#interactionLike']  #: User liked the feed item

InteractionRepost = t.Literal['app.bsky.feed.defs#interactionRepost']  #: User reposted the feed item

InteractionReply = t.Literal['app.bsky.feed.defs#interactionReply']  #: User replied to the feed item

InteractionQuote = t.Literal['app.bsky.feed.defs#interactionQuote']  #: User quoted the feed item

InteractionShare = t.Literal['app.bsky.feed.defs#interactionShare']  #: User shared the feed item



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2024 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t

from pydantic import Field

from atproto_client.models import string_formats

if t.TYPE_CHECKING:
    from atproto_client import models
from atproto_client.models import base


class Params(base.ParamsModelBase):
    """Parameters model for :obj:`app.bsky.feed.getPosts`."""

    uris: t.List[string_formats.AtUri] = Field(max_length=25)  #: List of post AT-URIs to return hydrated views for.


class ParamsDict(t.TypedDict):
    uris: t.List[string_formats.AtUri]  #: List of post AT-URIs to return hydrated views for.


class Response(base.ResponseModelBase):
    """Output data model for :obj:`app.bsky.feed.getPosts`."""

    posts: t.List['models.AppBskyFeedDefs.PostView']  #: Posts.



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2024 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing_extensions as te

#: Data raw data type.
Data: te.TypeAlias = bytes



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2024 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t

from atproto_client.models import base, string_formats


class Params(base.ParamsModelBase):
    """Parameters model for :obj:`com.atproto.sync.getHead`."""

    did: string_formats.Did  #: The DID of the repo.


class ParamsDict(t.TypedDict):
    did: string_formats.Did  #: The DID of the repo.


class Response(base.ResponseModelBase):
    """Output data model for :obj:`com.atproto.sync.getHead`."""

    root: string_formats.Cid  #: Root.



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2024 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t

import typing_extensions as te
from pydantic import Field

from atproto_client.models import string_formats

if t.TYPE_CHECKING:
    from atproto_client import models
    from atproto_client.models.unknown_type import UnknownType
from atproto_client.models import base


class Params(base.ParamsModelBase):
    """Parameters model for :obj:`app.bsky.notification.listNotifications`."""

    cursor: t.Optional[str] = None  #: Cursor.
    limit: t.Optional[int] = Field(default=50, ge=1, le=100)  #: Limit.
    priority: t.Optional[bool] = None  #: Priority.
    reasons: t.Optional[t.List[str]] = (
        None  #: Notification reasons to include in response. A reason that matches the reason property of #notification.
    )
    seen_at: t.Optional[string_formats.DateTime] = None  #: Seen at.


class ParamsDict(t.TypedDict):
    cursor: te.NotRequired[t.Optional[str]]  #: Cursor.
    limit: te.NotRequired[t.Optional[int]]  #: Limit.
    priority: te.NotRequired[t.Optional[bool]]  #: Priority.
    reasons: te.NotRequired[
        t.Optional[t.List[str]]
    ]  #: Notification reasons to include in response. A reason that matches the reason property of #notification.
    seen_at: te.NotRequired[t.Optional[string_formats.DateTime]]  #: Seen at.


class Response(base.ResponseModelBase):
    """Output data model for :obj:`app.bsky.notification.listNotifications`."""

    notifications: t.List['models.AppBskyNotificationListNotifications.Notification']  #: Notifications.
    cursor: t.Optional[str] = None  #: Cursor.
    priority: t.Optional[bool] = None  #: Priority.
    seen_at: t.Optional[string_formats.DateTime] = None  #: Seen at.


class Notification(base.ModelBase):
    """Definition model for :obj:`app.bsky.notification.listNotifications`."""

    author: 'models.AppBskyActorDefs.ProfileView'  #: Author.
    cid: string_formats.Cid  #: Cid.
    indexed_at: string_formats.DateTime  #: Indexed at.
    is_read: bool  #: Is read.
    reason: t.Union[
        t.Literal['like'],
        t.Literal['repost'],
        t.Literal['follow'],
        t.Literal['mention'],
        t.Literal['reply'],
        t.Literal['quote'],
        t.Literal['starterpack-joined'],
        t.Literal['verified'],
        t.Literal['unverified'],
        str,
    ]  #: Expected values are 'like', 'repost', 'follow', 'mention', 'reply', 'quote', 'starterpack-joined', 'verified', and 'unverified'.
    record: 'UnknownType'  #: Record.
    uri: string_formats.AtUri  #: Uri.
    labels: t.Optional[t.List['models.ComAtprotoLabelDefs.Label']] = None  #: Labels.
    reason_subject: t.Optional[string_formats.AtUri] = None  #: Reason subject.

    py_type: t.Literal['app.bsky.notification.listNotifications#notification'] = Field(
        default='app.bsky.notification.listNotifications#notification', alias='$type', frozen=True
    )



import typing as t

import typing_extensions as te
from pydantic import Field

if t.TYPE_CHECKING:
    from atproto_client import models
    from atproto_client.models import dot_dict

UnknownRecordType: te.TypeAlias = t.Union[
    'models.AppBskyActorProfile.Record',
    'models.AppBskyFeedGenerator.Record',
    'models.AppBskyFeedLike.Record',
    'models.AppBskyFeedPost.Record',
    'models.AppBskyFeedPostgate.Record',
    'models.AppBskyFeedRepost.Record',
    'models.AppBskyFeedThreadgate.Record',
    'models.AppBskyGraphBlock.Record',
    'models.AppBskyGraphFollow.Record',
    'models.AppBskyGraphList.Record',
    'models.AppBskyGraphListblock.Record',
    'models.AppBskyGraphListitem.Record',
    'models.AppBskyGraphStarterpack.Record',
    'models.AppBskyGraphVerification.Record',
    'models.AppBskyLabelerService.Record',
    'models.ChatBskyActorDeclaration.Record',
    'models.ComAtprotoLexiconSchema.Record',
]
UnknownRecordTypePydantic = te.Annotated[
    t.Union[
        'models.AppBskyActorProfile.Record',
        'models.AppBskyFeedGenerator.Record',
        'models.AppBskyFeedLike.Record',
        'models.AppBskyFeedPost.Record',
        'models.AppBskyFeedPostgate.Record',
        'models.AppBskyFeedRepost.Record',
        'models.AppBskyFeedThreadgate.Record',
        'models.AppBskyGraphBlock.Record',
        'models.AppBskyGraphFollow.Record',
        'models.AppBskyGraphList.Record',
        'models.AppBskyGraphListblock.Record',
        'models.AppBskyGraphListitem.Record',
        'models.AppBskyGraphStarterpack.Record',
        'models.AppBskyGraphVerification.Record',
        'models.AppBskyLabelerService.Record',
        'models.ChatBskyActorDeclaration.Record',
        'models.ComAtprotoLexiconSchema.Record',
    ],
    Field(discriminator='py_type'),
]
UnknownType: te.TypeAlias = t.Union[UnknownRecordTypePydantic, 'dot_dict.DotDictType']
UnknownInputType: te.TypeAlias = t.Union[UnknownType, t.Dict[str, t.Any]]



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2024 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t

import typing_extensions as te
from pydantic import Field

from atproto_client.models import string_formats

if t.TYPE_CHECKING:
    from atproto_client import models
from atproto_client.models import base


class Record(base.RecordModelBase):
    """Record model for :obj:`app.bsky.labeler.service`."""

    created_at: string_formats.DateTime  #: Created at.
    policies: 'models.AppBskyLabelerDefs.LabelerPolicies'  #: Policies.
    labels: t.Optional[
        te.Annotated[t.Union['models.ComAtprotoLabelDefs.SelfLabels'], Field(default=None, discriminator='py_type')]
    ] = None  #: Labels.
    reason_types: t.Optional[t.List['models.ComAtprotoModerationDefs.ReasonType']] = (
        None  #: The set of report reason 'codes' which are in-scope for this service to review and action. These usually align to policy categories. If not defined (distinct from empty array), all reason types are allowed.
    )
    subject_collections: t.Optional[t.List[string_formats.Nsid]] = (
        None  #: Set of record types (collection NSIDs) which can be reported to this service. If not defined (distinct from empty array), default is any record type.
    )
    subject_types: t.Optional[t.List['models.ComAtprotoModerationDefs.SubjectType']] = (
        None  #: The set of subject types (account, record, etc) this service accepts reports on.
    )

    py_type: t.Literal['app.bsky.labeler.service'] = Field(
        default='app.bsky.labeler.service', alias='$type', frozen=True
    )


class GetRecordResponse(base.SugarResponseModelBase):
    """Get record response for :obj:`models.AppBskyLabelerService.Record`."""

    uri: str  #: The URI of the record.
    value: 'models.AppBskyLabelerService.Record'  #: The record.
    cid: t.Optional[str] = None  #: The CID of the record.


class ListRecordsResponse(base.SugarResponseModelBase):
    """List records response for :obj:`models.AppBskyLabelerService.Record`."""

    records: t.Dict[str, 'models.AppBskyLabelerService.Record']  #: Map of URIs to records.
    cursor: t.Optional[str] = None  #: Next page cursor.


class CreateRecordResponse(base.SugarResponseModelBase):
    """Create record response for :obj:`models.AppBskyLabelerService.Record`."""

    uri: str  #: The URI of the record.
    cid: str  #: The CID of the record.



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2024 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t

import typing_extensions as te

if t.TYPE_CHECKING:
    from atproto_client.models.unknown_type import UnknownInputType, UnknownType
from atproto_client.models import base


class Data(base.DataModelBase):
    """Input data model for :obj:`com.atproto.identity.signPlcOperation`."""

    also_known_as: t.Optional[t.List[str]] = None  #: Also known as.
    rotation_keys: t.Optional[t.List[str]] = None  #: Rotation keys.
    services: t.Optional['UnknownInputType'] = None  #: Services.
    token: t.Optional[str] = None  #: A token received through com.atproto.identity.requestPlcOperationSignature.
    verification_methods: t.Optional['UnknownInputType'] = None  #: Verification methods.


class DataDict(t.TypedDict):
    also_known_as: te.NotRequired[t.Optional[t.List[str]]]  #: Also known as.
    rotation_keys: te.NotRequired[t.Optional[t.List[str]]]  #: Rotation keys.
    services: te.NotRequired[t.Optional['UnknownInputType']]  #: Services.
    token: te.NotRequired[
        t.Optional[str]
    ]  #: A token received through com.atproto.identity.requestPlcOperationSignature.
    verification_methods: te.NotRequired[t.Optional['UnknownInputType']]  #: Verification methods.


class Response(base.ResponseModelBase):
    """Output data model for :obj:`com.atproto.identity.signPlcOperation`."""

    operation: 'UnknownType'  #: A signed DID PLC operation.



import re
import typing as t
from collections import abc
from copy import deepcopy

import typing_extensions as te
from pydantic import GetJsonSchemaHandler
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import core_schema

from atproto_client.models.base import UnknownDict


def _convert_camel_case_to_snake_case(string: str) -> str:
    s = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', string)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s).lower()


def _convert_snake_case_to_camel_case(string: str) -> str:
    s = ''.join([w.capitalize() for w in string.split('_')])
    return s[0].lower() + s[1:]


def _is_snake_case(string: str) -> bool:
    return string == _convert_camel_case_to_snake_case(string)


def _convert_to_opposite_case(string: str) -> str:
    if _is_snake_case(string):
        return _convert_snake_case_to_camel_case(string)
    return _convert_camel_case_to_snake_case(string)


class DotDict(UnknownDict):
    """Dot notation for dictionaries.

    Note:
        If the record is out of the official lexicon, it`s impossible to deserialize it to a proper data model.
        Such models will fall back to dictionaries.
        All unknown "Union" types will also be caught as dicts.
        This class exists to provide an ability to use such fallbacks as “real” data models.

    Example:
        >>> test_data = {'a': 1, 'b': {'c': 2}, 'd': [{'e': 3}, 4, 5]}
        >>> model = DotDict(test_data)
        >>> assert isinstance(model, DotDict)
        >>> assert model.nonExistingField is None
        >>> assert model.a == 1
        >>> assert model['a'] == 1
        >>> assert model['b']['c'] == 2
        >>> assert model.b.c == 2
        >>> assert model.b['c'] == 2
        >>> assert model['b'].c == 2
        >>> assert model.d[0].e == 3
        >>> assert model['d'][0]['e'] == 3
        >>> assert model['d'][0].e == 3
        >>> assert model['d'][1] == 4
        >>> assert model['d'][2] == 5
        >>> model['d'][0]['e'] = 6
        >>> assert model['d'][0]['e'] == 6
        >>> assert DotDict(test_data) == DotDict(test_data)
        >>> assert model.to_dict() == test_data
    """

    def __init__(self, data: t.Dict[str, t.Any]) -> None:
        self._data = deepcopy(data)
        for k, v in self._data.items():
            self.__setitem__(k, v)

    def to_dict(self) -> t.Dict[str, t.Any]:
        """Unwrap DotDict to Python built-in dict."""
        return deepcopy(self._data)

    def __getitem__(self, item: str) -> t.Optional[t.Any]:
        value = self._data.get(item)
        if value is not None:
            return value

        return self._data.get(_convert_to_opposite_case(item))

    __getattr__ = __getitem__

    def __setitem__(self, key: str, value: t.Any) -> None:
        if key == '_data':
            super().__setattr__(key, value)
            return

        # we store the field in case that was firstly meet to not create duplicates
        if key not in self._data and _is_snake_case(key):
            key = _convert_snake_case_to_camel_case(key)

        self._data.__setitem__(key, DotDict.__convert(value))

    __setattr__ = __setitem__

    def __iter__(self) -> t.Iterator[t.Any]:
        yield from iter(self._data)

    def keys(self) -> abc.KeysView:
        return self._data.keys()

    def values(self) -> abc.ValuesView:
        return self._data.values()

    def items(self) -> abc.ItemsView:
        return self._data.items()

    def __len__(self) -> int:
        return len(self._data)

    def __contains__(self, item: t.Any) -> bool:
        return item in self._data

    def __eq__(self, other: t.Any) -> bool:
        if isinstance(other, DotDict):
            return self._data == other._data
        if isinstance(other, dict):
            return self._data == other

        raise NotImplementedError

    def __str__(self) -> str:
        return str(self._data)

    def __repr__(self) -> str:
        return repr(self._data)

    def __reduce_ex__(self, protocol: te.SupportsIndex):  # noqa: ANN204
        return self._data.__reduce_ex__(protocol)

    def __reduce__(self):  # noqa: ANN204
        return self._data.__reduce__()

    @staticmethod
    def __convert(obj: t.Any) -> t.Any:
        if isinstance(obj, dict):
            return DotDict(t.cast(t.Dict[str, t.Any], obj))
        if isinstance(obj, list):
            return [DotDict.__convert(v) for v in t.cast(t.List[t.Any], obj)]
        if isinstance(obj, set):
            return {DotDict.__convert(v) for v in t.cast(t.Set[t.Any], obj)}
        if isinstance(obj, tuple):
            return tuple(DotDict.__convert(v) for v in t.cast(t.Tuple[t.Any], obj))
        return obj


class _DotDictPydanticAnnotation:
    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: t.Any,
        _handler: t.Callable[[t.Any], core_schema.CoreSchema],
    ) -> core_schema.CoreSchema:
        """We return a pydantic_core.CoreSchema that behaves in the following ways below.

        * dicts will be parsed as `DotDict` instances with the int as the _data attribute
        * `DotDict` instances will be parsed as `DotDict` instances without any changes
        * Nothing else will pass validation
        * Serialization will always return just a dict
        """

        def validate_from_dict(value: t.Dict[str, t.Any]) -> DotDict:
            return DotDict(value)

        from_dict_schema = core_schema.chain_schema(
            [
                core_schema.dict_schema(),
                core_schema.no_info_plain_validator_function(validate_from_dict),
            ]
        )

        return core_schema.json_or_python_schema(
            json_schema=from_dict_schema,
            python_schema=core_schema.union_schema(
                [
                    # check if it's an instance first before doing any further work
                    core_schema.is_instance_schema(DotDict),
                    from_dict_schema,
                ]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(lambda instance: instance.to_dict()),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, _core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        # Use the same schema that would be used for `dict`
        return handler(core_schema.dict_schema())


DotDictType = te.Annotated[DotDict, _DotDictPydanticAnnotation]



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2024 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t

from pydantic import Field

from atproto_client.models import string_formats

if t.TYPE_CHECKING:
    from atproto_client.models.blob_ref import BlobRef
from atproto_client.models import base


class JobStatus(base.ModelBase):
    """Definition model for :obj:`app.bsky.video.defs`."""

    did: string_formats.Did  #: Did.
    job_id: str  #: Job id.
    state: t.Union[
        t.Literal['JOB_STATE_COMPLETED'], t.Literal['JOB_STATE_FAILED'], str
    ]  #: The state of the video processing job. All values not listed as a known value indicate that the job is in process.
    blob: t.Optional['BlobRef'] = None  #: Blob.
    error: t.Optional[str] = None  #: Error.
    message: t.Optional[str] = None  #: Message.
    progress: t.Optional[int] = Field(default=None, ge=0, le=100)  #: Progress within the current processing state.

    py_type: t.Literal['app.bsky.video.defs#jobStatus'] = Field(
        default='app.bsky.video.defs#jobStatus', alias='$type', frozen=True
    )



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2024 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t

import typing_extensions as te
from pydantic import Field

from atproto_client.models import string_formats

if t.TYPE_CHECKING:
    from atproto_client import models
from atproto_client.models import base


class Params(base.ParamsModelBase):
    """Parameters model for :obj:`com.atproto.sync.listReposByCollection`."""

    collection: string_formats.Nsid  #: Collection.
    cursor: t.Optional[str] = None  #: Cursor.
    limit: t.Optional[int] = Field(
        default=500, ge=1, le=2000
    )  #: Maximum size of response set. Recommend setting a large maximum (1000+) when enumerating large DID lists.


class ParamsDict(t.TypedDict):
    collection: string_formats.Nsid  #: Collection.
    cursor: te.NotRequired[t.Optional[str]]  #: Cursor.
    limit: te.NotRequired[
        t.Optional[int]
    ]  #: Maximum size of response set. Recommend setting a large maximum (1000+) when enumerating large DID lists.


class Response(base.ResponseModelBase):
    """Output data model for :obj:`com.atproto.sync.listReposByCollection`."""

    repos: t.List['models.ComAtprotoSyncListReposByCollection.Repo']  #: Repos.
    cursor: t.Optional[str] = None  #: Cursor.


class Repo(base.ModelBase):
    """Definition model for :obj:`com.atproto.sync.listReposByCollection`."""

    did: string_formats.Did  #: Did.

    py_type: t.Literal['com.atproto.sync.listReposByCollection#repo'] = Field(
        default='com.atproto.sync.listReposByCollection#repo', alias='$type', frozen=True
    )
